@@ -32,6 +32,8 @@
 #include "gg.h"
 #include "opt.h"
 
+#define	NREGVAR	24
+#define	REGBITS	((uint32)0xffffff)
 #define	P2R(p)	(Reg*)(p->reg)
 
 	void	addsplits(void);
@@ -128,6 +130,33 @@ setaddrs(Bits bit)
 	}
 }
 
+static char* regname[] = {
+	".R0",
+	".R1",
+	".R2",
+	".R3",
+	".R4",
+	".R5",
+	".R6",
+	".R7",
+	".R8",
+	".R9",
+	".R10",
+	".R11",
+	".R12",
+	".R13",
+	".R14",
+	".R15",
+	".F0",
+	".F1",
+	".F2",
+	".F3",
+	".F4",
+	".F5",
+	".F6",
+	".F7",
+};
+
 void
 regopt(Prog *firstp)
 {
@@ -136,7 +165,7 @@ regopt(Prog *firstp)
 	int i, z, nr;
 	uint32 vreg;
 	Bits bit;
-
+	
 	if(first == 0) {
 		fmtinstall('Q', Qconv);
 	}
@@ -164,7 +193,17 @@ regopt(Prog *firstp)
 	r1 = R;
 	firstr = R;
 	lastr = R;
-	nvar = 0;
+
+	/*
+	 * control flow is more complicated in generated go code
+	 * than in generated c code.  define pseudo-variables for
+	 * registers, so we have complete register usage information.
+	 */
+	nvar = NREGVAR;
+	memset(var, 0, NREGVAR*sizeof var[0]);
+	for(i=0; i<NREGVAR; i++)
+		var[i].sym = lookup(regname[i]);
+
 	regbits = RtoB(REGSP)|RtoB(REGLINK)|RtoB(REGPC);
 	for(z=0; z<BITS; z++) {
 		externs.b[z] = 0;
@@ -223,6 +262,16 @@ regopt(Prog *firstp)
 		bit = mkvar(r, &p->from);
 		for(z=0; z<BITS; z++)
 			r->use1.b[z] |= bit.b[z];
+		
+		/*
+		 * middle always read when present
+		 */
+		if(p->reg != NREG) {
+			if(p->from.type != D_FREG)
+				r->use1.b[0] |= RtoB(p->reg);
+			else
+				r->use1.b[0] |= FtoB(p->reg);
+		}
 
 		/*
 		 * right side depends on opcode
@@ -233,6 +282,68 @@ regopt(Prog *firstp)
 		default:
 			yyerror("reg: unknown op: %A", p->as);
 			break;
+		
+		/*
+		 * right side read
+		 */
+		case ATST:
+		case ATEQ:
+		case ACMP:
+		case ACMN:
+		case ACMPD:
+		case ACMPF:
+		rightread:
+			for(z=0; z<BITS; z++)
+				r->use2.b[z] |= bit.b[z];
+			break;
+			
+		/*
+		 * right side read or read+write, depending on middle
+		 *	ADD x, z => z += x
+		 *	ADD x, y, z  => z = x + y
+		 */
+		case AADD:
+		case AAND:
+		case AEOR:
+		case ASUB:
+		case ARSB:
+		case AADC:
+		case ASBC:
+		case ARSC:
+		case AORR:
+		case ABIC:
+		case ASLL:
+		case ASRL:
+		case ASRA:
+		case AMUL:
+		case AMULU:
+		case ADIV:
+		case AMOD:
+		case AMODU:
+		case ADIVU:
+			if(p->reg != NREG)
+				goto rightread;
+			// fall through
+
+		/*
+		 * right side read+write
+		 */
+		case AADDF:
+		case AADDD:
+		case ASUBF:
+		case ASUBD:
+		case AMULF:
+		case AMULD:
+		case ADIVF:
+		case ADIVD:
+		case AMULA:
+		case AMULAL:
+		case AMULALU:
+			for(z=0; z<BITS; z++) {
+				r->use2.b[z] |= bit.b[z];
+				r->set.b[z] |= bit.b[z];
+			}
+			break;
 
 		/*
 		 * right side write
@@ -240,11 +351,22 @@ regopt(Prog *firstp)
 		case ANOP:
 		case AMOVB:
 		case AMOVBU:
+		case AMOVD:
+		case AMOVDF:
+		case AMOVDW:
+		case AMOVF:
+		case AMOVFW:
 		case AMOVH:
 		case AMOVHU:
 		case AMOVW:
-		case AMOVF:
-		case AMOVD:
+		case AMOVWD:
+		case AMOVWF:
+		case AMVN:
+		case AMULL:
+		case AMULLU:
+			if((p->scond & C_SCOND) != C_SCOND_NONE)
+				for(z=0; z<BITS; z++)
+					r->use2.b[z] |= bit.b[z];
 			for(z=0; z<BITS; z++)
 				r->set.b[z] |= bit.b[z];
 			break;
@@ -396,6 +518,24 @@ loop2:
 		}
 	}
 
+	/*
+	 * pass 4.5
+	 * move register pseudo-variables into regu.
+	 */
+	for(r = firstr; r != R; r = r->link) {
+		r->regu = (r->refbehind.b[0] | r->set.b[0]) & REGBITS;
+
+		r->set.b[0] &= ~REGBITS;
+		r->use1.b[0] &= ~REGBITS;
+		r->use2.b[0] &= ~REGBITS;
+		r->refbehind.b[0] &= ~REGBITS;
+		r->refahead.b[0] &= ~REGBITS;
+		r->calbehind.b[0] &= ~REGBITS;
+		r->calahead.b[0] &= ~REGBITS;
+		r->regdiff.b[0] &= ~REGBITS;
+		r->act.b[0] &= ~REGBITS;
+	}
+
 	/*
 	 * pass 5
 	 * isolate regions
@@ -589,13 +729,19 @@ addsplits(void)
 void
 addmove(Reg *r, int bn, int rn, int f)
 {
-	Prog *p, *p1;
+	Prog *p, *p1, *p2;
 	Adr *a;
 	Var *v;
 
 	p1 = mal(sizeof(*p1));
 	*p1 = zprog;
 	p = r->prog;
+	
+	// If there's a stack fixup coming (after BL newproc or BL deferproc),
+	// delay the load until after the fixup.
+	p2 = p->link;
+	if(p2 && p2->as == AMOVW && p2->from.type == D_CONST && p2->from.reg == REGSP && p2->to.reg == REGSP && p2->to.type == D_REG)
+		p = p2;
 
 	p1->link = p->link;
 	p->link = p1;
@@ -606,6 +752,7 @@ addmove(Reg *r, int bn, int rn, int f)
 	a = &p1->to;
 	a->sym = v->sym;
 	a->name = v->name;
+	a->node = v->node;
 	a->offset = v->offset;
 	a->etype = v->etype;
 	a->type = D_OREG;
@@ -624,6 +771,7 @@ addmove(Reg *r, int bn, int rn, int f)
 		break;
 	case TBOOL:
 	case TUINT8:
+//print("movbu %E %d %S\n", v->etype, bn, v->sym);
 		p1->as = AMOVBU;
 		break;
 	case TINT16:
@@ -697,9 +845,6 @@ mkvar(Reg *r, Adr *a)
 	n = D_NONE;
 
 	flag = 0;
-	if(a->pun)
-		flag = 1;
-
 	switch(t) {
 	default:
 		print("type %d %d %D\n", t, a->name, a);
@@ -715,21 +860,40 @@ mkvar(Reg *r, Adr *a)
 		goto onereg;
 
 	case D_REGREG:
+		bit = zbits;
 		if(a->offset != NREG)
-			r->regu |= RtoB(a->offset);
-		goto onereg;
+			bit.b[0] |= RtoB(a->offset);
+		if(a->reg != NREG)
+			bit.b[0] |= RtoB(a->reg);
+		return bit;
 
 	case D_REG:
 	case D_SHIFT:
-	case D_OREG:
 	onereg:
-		if(a->reg != NREG)
-			r->regu |= RtoB(a->reg);
+		if(a->reg != NREG) {
+			bit = zbits;
+			bit.b[0] = RtoB(a->reg);
+			return bit;
+		}
+		break;
+
+	case D_OREG:
+		if(a->reg != NREG) {
+			if(a == &r->prog->from)
+				r->use1.b[0] |= RtoB(a->reg);
+			else
+				r->use2.b[0] |= RtoB(a->reg);
+			if(r->prog->scond & (C_PBIT|C_WBIT))
+				r->set.b[0] |= RtoB(a->reg);
+		}
 		break;
 
 	case D_FREG:
-		if(a->reg != NREG)
-			r->regu |= FtoB(a->reg);
+		if(a->reg != NREG) {
+			bit = zbits;
+			bit.b[0] = FtoB(a->reg);
+			return bit;
+		}
 		break;
 	}
 
@@ -763,7 +927,7 @@ mkvar(Reg *r, Adr *a)
 				if(!flag)
 					return blsh(i);
 
-			// if they overlaps, disable both
+			// if they overlap, disable both
 			if(overlap(v->offset, v->width, o, w)) {
 				v->addr = 1;
 				flag = 1;
@@ -787,6 +951,7 @@ mkvar(Reg *r, Adr *a)
 
 	i = nvar;
 	nvar++;
+//print("var %d %E %D %S\n", i, et, a, s);
 	v = var+i;
 	v->sym = s;
 	v->offset = o;
@@ -795,9 +960,10 @@ mkvar(Reg *r, Adr *a)
 	v->etype = et;
 	v->width = w;
 	v->addr = flag;		// funny punning
-
+	v->node = a->node;
+	
 	if(debug['R'])
-		print("bit=%2d et=%E pun=%d %D\n", i, et, flag, a);
+		print("bit=%2d et=%2d w=%d+%d %S %D flag=%d\n", i, et, o, w, s, a, v->addr);
 
 	bit = blsh(i);
 	if(n == D_EXTERN || n == D_STATIC)
@@ -1287,10 +1453,6 @@ paint3(Reg *r, int bn, int32 rb, int rn)
 void
 addreg(Adr *a, int rn)
 {
-
-	if(a->type == D_CONST)
-		fatal("addreg: cant do this %D %d\n", a, rn);
-
 	a->sym = 0;
 	a->name = D_NONE;
 	a->type = D_REG;
@@ -1311,8 +1473,7 @@ addreg(Adr *a, int rn)
 int32
 RtoB(int r)
 {
-
-	if(r < 2 || r >= REGTMP-2)	// excluded R9 and R10 for m and g
+	if(r >= REGTMP-2)	// excluded R9 and R10 for m and g
 		return 0;
 	return 1L << r;
 }
@@ -1365,6 +1526,7 @@ noreturn(Prog *p)
 		symlist[1] = pkglookup("panicslice", runtimepkg);
 		symlist[2] = pkglookup("throwinit", runtimepkg);
 		symlist[3] = pkglookup("panic", runtimepkg);
+		symlist[4] = pkglookup("panicwrap", runtimepkg);
 	}
 
 	s = p->to.sym;
@@ -1395,27 +1557,27 @@ dumpone(Reg *r)
 			r->regdiff.b[z] |
 			r->act.b[z] |
 				0;
-//	if(bany(&bit)) {
-//		print("\t");
-//		if(bany(&r->set))
-//			print(" s:%Q", r->set);
-//		if(bany(&r->use1))
-//			print(" u1:%Q", r->use1);
-//		if(bany(&r->use2))
-//			print(" u2:%Q", r->use2);
-//		if(bany(&r->refbehind))
-//			print(" rb:%Q ", r->refbehind);
-//		if(bany(&r->refahead))
-//			print(" ra:%Q ", r->refahead);
-//		if(bany(&r->calbehind))
-//			print("cb:%Q ", r->calbehind);
-//		if(bany(&r->calahead))
-//			print(" ca:%Q ", r->calahead);
-//		if(bany(&r->regdiff))
-//			print(" d:%Q ", r->regdiff);
-//		if(bany(&r->act))
-//			print(" a:%Q ", r->act);
-//	}
+	if(bany(&bit)) {
+		print("\t");
+		if(bany(&r->set))
+			print(" s:%Q", r->set);
+		if(bany(&r->use1))
+			print(" u1:%Q", r->use1);
+		if(bany(&r->use2))
+			print(" u2:%Q", r->use2);
+		if(bany(&r->refbehind))
+			print(" rb:%Q ", r->refbehind);
+		if(bany(&r->refahead))
+			print(" ra:%Q ", r->refahead);
+		if(bany(&r->calbehind))
+			print("cb:%Q ", r->calbehind);
+		if(bany(&r->calahead))
+			print(" ca:%Q ", r->calahead);
+		if(bany(&r->regdiff))
+			print(" d:%Q ", r->regdiff);
+		if(bany(&r->act))
+			print(" a:%Q ", r->act);
+	}
 	print("\n");
 }
 
