@@ -83,9 +83,6 @@ enum {
 	ElfStrText,
 	ElfStrData,
 	ElfStrBss,
-	ElfStrGosymcounts,
-	ElfStrGosymtab,
-	ElfStrGopclntab,
 	ElfStrShstrtab,
 	ElfStrSymtab,
 	ElfStrStrtab,
@@ -104,6 +101,9 @@ needlib(char *name)
 	char *p;
 	Sym *s;
 
+	if(*name == '\0')
+		return 0;
+
 	/* reuse hash code in symbol table */
 	p = smprint(".dynlib.%s", name);
 	s = lookup(p, 0);
@@ -519,7 +519,7 @@ doelf(void)
 
 	/* predefine strings we need for section headers */
 	shstrtab = lookup(".shstrtab", 0);
-	shstrtab->type = SELFDATA;
+	shstrtab->type = SELFROSECT;
 	shstrtab->reachable = 1;
 
 	elfstr[ElfStrEmpty] = addstring(shstrtab, "");
@@ -528,10 +528,11 @@ doelf(void)
 	elfstr[ElfStrBss] = addstring(shstrtab, ".bss");
 	addstring(shstrtab, ".elfdata");
 	addstring(shstrtab, ".rodata");
+	addstring(shstrtab, ".gosymtab");
+	addstring(shstrtab, ".gopclntab");
 	if(!debug['s']) {
-		elfstr[ElfStrGosymcounts] = addstring(shstrtab, ".gosymcounts");
-		elfstr[ElfStrGosymtab] = addstring(shstrtab, ".gosymtab");
-		elfstr[ElfStrGopclntab] = addstring(shstrtab, ".gopclntab");
+		elfstr[ElfStrSymtab] = addstring(shstrtab, ".symtab");
+		elfstr[ElfStrStrtab] = addstring(shstrtab, ".strtab");
 		dwarfaddshstrings(shstrtab);
 	}
 	elfstr[ElfStrShstrtab] = addstring(shstrtab, ".shstrtab");
@@ -550,21 +551,16 @@ doelf(void)
 		elfstr[ElfStrGnuVersion] = addstring(shstrtab, ".gnu.version");
 		elfstr[ElfStrGnuVersionR] = addstring(shstrtab, ".gnu.version_r");
 
-		/* interpreter string */
-		s = lookup(".interp", 0);
-		s->reachable = 1;
-		s->type = SELFDATA;
-
 		/* dynamic symbol table - first entry all zeros */
 		s = lookup(".dynsym", 0);
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 		s->reachable = 1;
 		s->size += ELF32SYMSIZE;
 
 		/* dynamic string table */
 		s = lookup(".dynstr", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 		if(s->size == 0)
 			addstring(s, "");
 		dynstr = s;
@@ -572,45 +568,45 @@ doelf(void)
 		/* relocation table */
 		s = lookup(".rel", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 
 		/* global offset table */
 		s = lookup(".got", 0);
 		s->reachable = 1;
-		s->type = SDATA;	// writable, so not SELFDATA
+		s->type = SELFSECT; // writable
 		
 		/* hash */
 		s = lookup(".hash", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 
 		/* got.plt */
 		s = lookup(".got.plt", 0);
 		s->reachable = 1;
-		s->type = SDATA;	// writable, so not SELFDATA
+		s->type = SELFSECT; // writable
 		
 		s = lookup(".plt", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 
 		s = lookup(".rel.plt", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 		
 		s = lookup(".gnu.version", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 		
 		s = lookup(".gnu.version_r", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 
 		elfsetupplt();
 
 		/* define dynamic elf table */
 		s = lookup(".dynamic", 0);
 		s->reachable = 1;
-		s->type = SELFDATA;
+		s->type = SELFROSECT;
 
 		/*
 		 * .dynamic table
@@ -637,8 +633,11 @@ doelf(void)
 void
 shsym(Elf64_Shdr *sh, Sym *s)
 {
-	sh->addr = symaddr(s);
-	sh->off = datoff(sh->addr);
+	vlong addr;
+	addr = symaddr(s);
+	if(sh->flags&SHF_ALLOC)
+		sh->addr = addr;
+	sh->off = datoff(addr);
 	sh->size = s->size;
 }
 
@@ -664,6 +663,7 @@ asmb(void)
 	ElfShdr *sh;
 	Section *sect;
 	Sym *sym;
+	int o;
 	int i;
 
 	if(debug['v'])
@@ -671,19 +671,20 @@ asmb(void)
 	Bflush(&bso);
 
 	sect = segtext.sect;
-	seek(cout, sect->vaddr - segtext.vaddr + segtext.fileoff, 0);
+	cseek(sect->vaddr - segtext.vaddr + segtext.fileoff);
 	codeblk(sect->vaddr, sect->len);
 
-	/* output read-only data in text segment */
-	sect = segtext.sect->next;
-	seek(cout, sect->vaddr - segtext.vaddr + segtext.fileoff, 0);
-	datblk(sect->vaddr, sect->len);
+	/* output read-only data in text segment (rodata, gosymtab and pclntab) */
+	for(sect = sect->next; sect != nil; sect = sect->next) {
+		cseek(sect->vaddr - segtext.vaddr + segtext.fileoff);
+		datblk(sect->vaddr, sect->len);
+	}
 
 	if(debug['v'])
 		Bprint(&bso, "%5.2f datblk\n", cputime());
 	Bflush(&bso);
 
-	seek(cout, segdata.fileoff, 0);
+	cseek(segdata.fileoff);
 	datblk(segdata.vaddr, segdata.filelen);
 
 	machlink = 0;
@@ -693,7 +694,7 @@ asmb(void)
 	if(iself) {
 		/* index of elf text section; needed by asmelfsym, double-checked below */
 		/* !debug['d'] causes extra sections before the .text section */
-		elftextsh = 1;
+		elftextsh = 2;
 		if(!debug['d']) {
 			elftextsh += 10;
 			if(elfverneed)
@@ -715,10 +716,10 @@ asmb(void)
 			if(iself)
 				goto Elfsym;
 		case Hgarbunix:
-			seek(cout, rnd(HEADR+segtext.filelen, 8192)+segdata.filelen, 0);
+			symo = rnd(HEADR+segtext.filelen, 8192)+segdata.filelen;
 			break;
 		case Hunixcoff:
-			seek(cout, rnd(HEADR+segtext.filelen, INITRND)+segdata.filelen, 0);
+			symo = rnd(HEADR+segtext.filelen, INITRND)+segdata.filelen;
 			break;
 		case Hplan9x32:
 			symo = HEADR+segtext.filelen+segdata.filelen;
@@ -736,38 +737,50 @@ asmb(void)
 			symo = rnd(symo, INITRND);
 			break;
 		case Hwindows:
-			// TODO(brainman): not sure what symo meant to be, but it is not used for Windows PE for now anyway
 			symo = rnd(HEADR+segtext.filelen, PEFILEALIGN)+segdata.filelen;
 			symo = rnd(symo, PEFILEALIGN);
 			break;
 		}
-		if(!debug['s']) {
-			seek(cout, symo, 0);
-			
-			if(HEADTYPE == Hplan9x32) {
-				asmplan9sym();
+		cseek(symo);
+		switch(HEADTYPE) {
+		default:
+			if(iself) {
+				if(debug['v'])
+				       Bprint(&bso, "%5.2f elfsym\n", cputime());
+				asmelfsym();
 				cflush();
-				
-				sym = lookup("pclntab", 0);
-				if(sym != nil) {
-					lcsize = sym->np;
-					for(i=0; i < lcsize; i++)
-						cput(sym->p[i]);
-					
-					cflush();
-				}
-				
-			} else if(HEADTYPE != Hwindows) {
+				cwrite(elfstrdat, elfstrsize);
+
 				if(debug['v'])
 					Bprint(&bso, "%5.2f dwarf\n", cputime());
 				dwarfemitdebugsections();
 			}
+			break;
+		case Hplan9x32:
+			asmplan9sym();
+			cflush();
+
+			sym = lookup("pclntab", 0);
+			if(sym != nil) {
+				lcsize = sym->np;
+				for(i=0; i < lcsize; i++)
+					cput(sym->p[i]);
+				
+				cflush();
+			}
+			break;
+		case Hdarwin:
+		case Hwindows:
+			if(debug['v'])
+				Bprint(&bso, "%5.2f dwarf\n", cputime());
+			dwarfemitdebugsections();
+			break;
 		}
 	}
 	if(debug['v'])
 		Bprint(&bso, "%5.2f headr\n", cputime());
 	Bflush(&bso);
-	seek(cout, 0L, 0);
+	cseek(0L);
 	switch(HEADTYPE) {
 	default:
 		if(iself)
@@ -903,10 +916,6 @@ asmb(void)
 		break;
 
 	Elfput:
-		/* elf 386 */
-		if(HEADTYPE == Htiny)
-			debug['d'] = 1;
-
 		eh = getElfEhdr();
 		startva = INITTEXT - HEADR;
 
@@ -922,6 +931,17 @@ asmb(void)
 		pph->paddr = INITTEXT - HEADR + pph->off;
 		pph->align = INITRND;
 
+		/*
+		 * PHDR must be in a loaded segment. Adjust the text
+		 * segment boundaries downwards to include it.
+		 */
+		o = segtext.vaddr - pph->vaddr;
+		segtext.vaddr -= o;
+		segtext.len += o;
+		o = segtext.fileoff - pph->off;
+		segtext.fileoff -= o;
+		segtext.filelen += o;
+
 		if(!debug['d']) {
 			/* interpreter */
 			sh = newElfShdr(elfstr[ElfStrInterp]);
@@ -1062,6 +1082,11 @@ asmb(void)
 		ph->flags = PF_W+PF_R;
 		ph->align = 4;
 
+		sh = newElfShstrtab(elfstr[ElfStrShstrtab]);
+		sh->type = SHT_STRTAB;
+		sh->addralign = 1;
+		shsym(sh, lookup(".shstrtab", 0));
+
 		if(elftextsh != eh->shnum)
 			diag("elftextsh = %d, want %d", elftextsh, eh->shnum);
 		for(sect=segtext.sect; sect!=nil; sect=sect->next)
@@ -1070,26 +1095,23 @@ asmb(void)
 			elfshbits(sect);
 
 		if (!debug['s']) {
-			sh = newElfShdr(elfstr[ElfStrGosymtab]);
-			sh->type = SHT_PROGBITS;
-			sh->flags = SHF_ALLOC;
-			sh->addralign = 1;
-			shsym(sh, lookup("symtab", 0));
+			sh = newElfShdr(elfstr[ElfStrSymtab]);
+			sh->type = SHT_SYMTAB;
+			sh->off = symo;
+			sh->size = symsize;
+			sh->addralign = 4;
+			sh->entsize = 16;
+			sh->link = eh->shnum;	// link to strtab
 
-			sh = newElfShdr(elfstr[ElfStrGopclntab]);
-			sh->type = SHT_PROGBITS;
-			sh->flags = SHF_ALLOC;
+			sh = newElfShdr(elfstr[ElfStrStrtab]);
+			sh->type = SHT_STRTAB;
+			sh->off = symo+symsize;
+			sh->size = elfstrsize;
 			sh->addralign = 1;
-			shsym(sh, lookup("pclntab", 0));
 
 			dwarfaddelfheaders();
 		}
 
-		sh = newElfShstrtab(elfstr[ElfStrShstrtab]);
-		sh->type = SHT_STRTAB;
-		sh->addralign = 1;
-		shsym(sh, lookup(".shstrtab", 0));
-
 		/* Main header */
 		eh->ident[EI_MAG0] = '\177';
 		eh->ident[EI_MAG1] = 'E';
@@ -1114,13 +1136,13 @@ asmb(void)
 			pph->memsz = pph->filesz;
 		}
 
-		seek(cout, 0, 0);
+		cseek(0);
 		a = 0;
 		a += elfwritehdr();
 		a += elfwritephdrs();
 		a += elfwriteshdrs();
 		cflush();
-		if(a+elfwriteinterp() > ELFRESERVE)
+		if(a+elfwriteinterp() > ELFRESERVE)	
 			diag("ELFRESERVE too small: %d > %d", a, ELFRESERVE);
 		break;
 
@@ -1142,25 +1164,6 @@ s8put(char *n)
 		cput(name[i]);
 }
 
-void
-cflush(void)
-{
-	int n;
-
-	n = sizeof(buf.cbuf) - cbc;
-	if(n)
-		ewrite(cout, buf.cbuf, n);
-	cbp = buf.cbuf;
-	cbc = sizeof(buf.cbuf);
-}
-
-/* Current position in file */
-vlong
-cpos(void)
-{
-	return seek(cout, 0, 1) + sizeof(buf.cbuf) - cbc;
-}
-
 int32
 rnd(int32 v, int32 r)
 {
@@ -1195,7 +1198,7 @@ genasmsym(void (*put)(Sym*, char*, int, vlong, vlong, int, Sym*))
 			case SCONST:
 			case SRODATA:
 			case SDATA:
-			case SELFDATA:
+			case SELFROSECT:
 			case SMACHO:
 			case SMACHOGOT:
 			case STYPE:
@@ -1245,6 +1248,6 @@ genasmsym(void (*put)(Sym*, char*, int, vlong, vlong, int, Sym*))
 				put(nil, a->asym->name, 'p', a->aoffset, 0, 0, a->gotype);
 	}
 	if(debug['v'] || debug['n'])
-		Bprint(&bso, "symsize = %uld\n", symsize);
+		Bprint(&bso, "symsize = %d\n", symsize);
 	Bflush(&bso);
 }
